-module(mock).
-export([start/0, stop/0, call/1, next/0, answer/1]).
-include_lib("eunit/include/eunit.hrl").
-define(TIMEOUT, 2000).


start() ->
    register(mock, spawn(fun loop/0)).

stop() ->
    mock ! stop.

ensure_started() ->
    case whereis(mock) of
        undefined ->
            start(),
            ok;
        _ ->
            ok
    end.

call(Call) ->
    ensure_started(),
    mock ! {call, self(), Call},
    receive
        {answer, Answer} ->
            Answer;
        timeout ->
            throw(mock_call_timeout)
    end.

next() ->
    ensure_started(),
    mock ! {next, self()},
    receive
        {call, Call} ->
            Call;
        timeout ->
            mock_next_timeout
    end.

answer(Answer) ->
    ensure_started(),
    mock ! {answer, Answer},
    ok.

loop() ->
    receive
        {call, CallOrigin, Call} ->
            receive
                {next, NextOrigin} ->
                    NextOrigin ! {call, Call},
                    receive
                        {answer, Answer} ->
                            CallOrigin ! {answer, Answer}
                    after
                        ?TIMEOUT ->
                            CallOrigin ! timeout
                    end;
                {answer, Answer} ->
                    CallOrigin ! {answer, Answer}
            after
                ?TIMEOUT ->
                    CallOrigin ! timeout
            end,
            loop();

        {next, NextOrigin} ->
            receive
                {call, CallOrigin, Call} ->
                    NextOrigin ! {call, Call},
                    receive
                        {answer, Answer} ->
                            CallOrigin ! {answer, Answer}
                    after
                        ?TIMEOUT ->
                            CallOrigin ! timeout
                    end
            after
                ?TIMEOUT ->
                    NextOrigin ! timeout
            end,
            loop();

        {answer, Answer} ->
            receive
                {call, CallOrigin, _Call} ->
                    CallOrigin ! {answer, Answer}
            after
                ?TIMEOUT -> ok
            end,
            loop();

        stop ->
            exit(normal)
    end.


usual_call_next_answer_test() ->
    Self = self(),
    spawn(fun() -> Self ! {result, mock:call(hello)} end),
    case mock:next() of
        hello ->
            mock:answer(world);
        Other ->
            throw({unexpected_next, Other})
    end,
    receive
        {result, Result} ->
            world = Result
    after 1 ->
            throw(timeout_before_answer)
    end.

next_before_call_test() ->
    Self = self(),
    spawn(fun() ->
              receive after 1 -> ok end,
              Self ! {result, mock:call(hello)}
          end),
    case mock:next() of
        hello ->
            mock:answer(world);
        Other ->
            throw({unexpected_next, Other})
    end,
    receive
        {result, Result} ->
            world = Result
    after 2 ->
            throw(timeout_before_answer)
    end.

answer_without_next_test() ->
    Self = self(),
    spawn(fun() ->
              Self ! {result, mock:call(hello)}
          end),
    receive after 1 -> ok end,
    mock:answer(world),
    receive
        {result, Result} ->
            world = Result
    after 1 ->
            throw(timeout_before_answer)
    end.

answer_before_call_test() ->
    mock:answer(world),
    ?assertEqual(world, mock:call(hello)).

call_timeout_test() ->
    ?assertEqual(mock_call_timeout, catch mock:call(hello)).
