-module(mup_config).
-behaviour(gen_server).
-export([start_link/0, reconfig/0, get_server/1, get_plugin/1,
         get_server_names/0, get_plugin_names/0, add_term/1]).
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).
-include("mup.hrl").
-include_lib("eunit/include/eunit.hrl").


-record(state,
        {terms=[]}).


start_link() ->
    gen_server:start_link({local, ?MODULE}, ?MODULE, [], []).

reconfig() ->
    gen_server:call(mup_config, reconfig).

get_server(Name) ->
    gen_server:call(mup_config, {get_config, server, Name}).

get_plugin(Name) ->
    gen_server:call(mup_config, {get_config, plugin, Name}).

get_server_names() ->
    gen_server:call(mup_config, {get_names, server}).

get_plugin_names() ->
    gen_server:call(mup_config, {get_names, plugin}).

add_term(Term) ->
    % For testing.
    gen_server:call(mup_config, {add_term, Term}).


init(_Args) ->
    {ok, #state{}}.

handle_call(reconfig, _From, State) ->
    case application:get_env(mup, config_path) of
        undefined ->
            {reply, {error, config_path_undefined}, State};
        {ok, Path} ->
            case file:consult(Path) of
                {ok, Terms} ->
                    {reply, ok, State#state{terms=Terms}};
                Other ->
                    {reply, Other, State}
            end
    end;

handle_call({get_config, Type, Name}, _From, State) ->
    {reply, find_config(Type, Name, State#state.terms), State};

handle_call({get_names, Type}, _From, State) ->
    {reply, collect_names(Type, State#state.terms), State};

handle_call({add_term, Term}, _From, State) ->
    % Append to the end to make testing more logical.
    {reply, ok, State#state{terms=State#state.terms ++ [Term]}}.

handle_cast(_Request, State) ->
    {noreply, State}.

handle_info(_Request, State) ->
    {noreply, State}.

terminate(_Rason, _State) ->
    ignored.

code_change(_OldVsn, State, _Extra) ->
    {ok, State}.


find_config(server, Name, [{server, Name, _Options}=Term|_]) ->
    {ok, Term};
find_config(plugin, Name, [{plugin, Name, _Module, _Options}=Term|_]) ->
    {ok, Term};
find_config(Type, Name, [_Term|Rest]) ->
    find_config(Type, Name, Rest);
find_config(_Type, _Name, []) ->
    undefined.

collect_names(server, [{server, Name, _Options}|Rest]) ->
    [Name|collect_names(server, Rest)];
collect_names(plugin, [{plugin, Name, _Module, _Options}|Rest]) ->
    [Name|collect_names(plugin, Rest)];
collect_names(Type, [_Term|Rest]) ->
    collect_names(Type, Rest);
collect_names(_Type, []) ->
    [].


mup_config_test_() ->
    {foreach,
     fun () ->
        {ok, Pid} = mup_config:start_link(),
        mup_config:add_term({server, server1, [options]}),
        mup_config:add_term({plugin, plugin1, the_module, [options]}),
        mup_config:add_term({server, server2, [options]}),
        mup_config:add_term({plugin, plugin2, the_module, [options]}),
        Pid
     end,
     fun (Pid) -> mup_util:sync_stop(Pid) end,
     [fun get_server_subtest/0,
      fun get_server_names_subtest/0,
      fun get_plugin_subtest/0,
      fun get_plugin_names_subtest/0]}.

get_server_subtest() ->
    ?assertEqual({ok, {server, server1, [options]}}, get_server(server1)),
    ?assertEqual({ok, {server, server2, [options]}}, get_server(server2)).

get_server_names_subtest() ->
    ?assertEqual([server1, server2], get_server_names()).

get_plugin_subtest() ->
    ?assertEqual({ok, {plugin, plugin1, the_module, [options]}},
                 get_plugin(plugin1)),
    ?assertEqual({ok, {plugin, plugin2, the_module, [options]}},
                 get_plugin(plugin2)).

get_plugin_names_subtest() ->
    ?assertEqual([plugin1, plugin2], get_plugin_names()).


reconfig_test_() ->
    {setup,
     fun () ->
        {ok, Pid} = mup_config:start_link(),
        TmpConfigPath = ?cmd("mktemp"),
        OldConfigPath = application:get_env(mup, config_path),
        {Pid, TmpConfigPath, OldConfigPath}
     end,
     fun ({Pid, TmpConfigPath, OldConfigPath}) ->
        application:set_env(mup, config_path, OldConfigPath),
        file:delete(TmpConfigPath),
        mup_util:sync_stop(Pid)
     end,
     fun ({_Pid, TmpConfigPath, _OldConfigPath}) -> [fun () ->
        application:set_env(mup, config_path, TmpConfigPath),
        file:write_file(TmpConfigPath, "{server, server_name, [options]}."),
        ?assertEqual(ok, mup_config:reconfig()),
        ?assertEqual({ok, {server, server_name, [options]}},
                     mup_config:get_server(server_name))
     end] end}.
