-module(mup_launchpad).
-behaviour(gen_server).
-export([start_link/1]).
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).
-include("mup.hrl").
-include_lib("eunit/include/eunit.hrl").

% TODO:
% - Show new bugs up for review
% - Show bugs that got out of review
% - Show new bugs filed
%   Bug #123454 opened by jkakar: The summary <review> <...>
% - Show bugs that got Fix Committed (not sure)

-record(state,
        {plugin}).

start_link(Name) ->
    gen_server:start_link(?MODULE, Name, []).

init(Name) ->
    {ok, Plugin} = mup_config:get_plugin(Name),
    {ok, #state{plugin=Plugin}}.

handle_call(_Request, _From, State) ->
    {reply, ok, State}.

handle_cast({handle_message, #message{command='PRIVMSG'}=Message}, State) ->
    ShouldHandle = should_handle_message(Message, State),
    if
        ShouldHandle ->
            case parse_bug_numbers(Message#message.text) of
                [] ->
                    ok;
                Bugs ->
                    spawn(fun () -> display_bugs(Bugs, Message, State) end)
            end;
        true ->
            ok
    end,
    {noreply, State};

handle_cast({handle_message, _Message}, State) ->
    {noreply, State};

handle_cast(reconfig, State) ->
    case mup_config:get_plugin((State#state.plugin)#plugin.name) of
        {ok, Plugin} ->
            {noreply, State#state{plugin=Plugin}};
        undefined ->
            {stop, normal, State}
    end.

handle_info(_Info, State) ->
    {noreply, State}.

terminate(_Reason, _State) ->
    ok.

code_change(_OldVsn, State, _Extra) ->
    {ok, State}.


should_handle_message(Message, State) ->
    HandledChannels = get_option(channels, State),
    (HandledChannels == undefined)
    orelse
    lists:member({(Message#message.server)#server.name,
                  lists:nth(1, Message#message.params)},
                 HandledChannels).


parse_bug_numbers([$#|Rest]) ->
    parse_bug_number(Rest, 10000);
parse_bug_numbers([B,$u,$g,$ |Rest]) when (B == $b) or (B == $B) ->
    parse_bug_number(Rest, 1);
parse_bug_numbers([$/,$+,$b,$u,$g,$/|Rest]) ->
    parse_bug_number(Rest, 1);
parse_bug_numbers([$/,$b,$u,$g,$s,$/|Rest]) ->
    parse_bug_number(Rest, 1);
parse_bug_numbers([_|Rest]) ->
    parse_bug_numbers(Rest);
parse_bug_numbers([]) ->
    [].

parse_bug_number([$ |Text], MinNumber) ->
    parse_bug_number(Text, MinNumber);
parse_bug_number([$#|Text], MinNumber) ->
    parse_bug_number(Text, MinNumber);
parse_bug_number(Text, MinNumber) ->
    case string:to_integer(Text) of
        {error, _Error} ->
            parse_bug_numbers(Text);
        {Number, AfterNumber} ->
            if
                Number >= MinNumber ->
                    [integer_to_list(Number)|parse_bug_numbers(AfterNumber)];
                true ->
                    parse_bug_numbers(AfterNumber)
            end
    end.


parse_bug_numbers_test_() ->
    [?_assertEqual(["123"], parse_bug_numbers("bug 123")),
     ?_assertEqual(["123"], parse_bug_numbers("bug #123")),
     ?_assertEqual(["123"], parse_bug_numbers("Bug #123")),
     ?_assertEqual(["123"], parse_bug_numbers("bug  ##123")),
     ?_assertEqual(["123"], parse_bug_numbers("One bug #123 is enough")),
     ?_assertEqual(["123", "456"], parse_bug_numbers("bug 123 bug 456")),
     ?_assertEqual([], parse_bug_numbers("bug #hah")),
     ?_assertEqual(["10000"], parse_bug_numbers("#10000")),
     ?_assertEqual(["10000"], parse_bug_numbers("One #10000 is enough")),
     ?_assertEqual([], parse_bug_numbers("One 10000 is enough")),
     ?_assertEqual(["123"], parse_bug_numbers("/+bug/123")),
     ?_assertEqual(["456"], parse_bug_numbers("123/+bug/456/789")),
     ?_assertEqual(["456"], parse_bug_numbers("123/bugs/456/789"))
    ].


display_bugs([Number|Numbers], Message, State) ->
    Bug = perform_launchpad_request(["/bugs/", Number], State),
    AllowPrivate = get_option(private, false, State),
    IsPrivate = get_object_value("private", Bug),
    if
        AllowPrivate or not IsPrivate ->
            display_bug(Bug, Number, Message, State);
        true ->
            ok
    end,
    display_bugs(Numbers, Message, State);
display_bugs([], _Message, _State) ->
    ok.

display_bug(Bug, Number, Message, State) ->
    BugTasksUrl = get_object_value("bug_tasks_collection_link", Bug),
    BugTasks = perform_launchpad_request(BugTasksUrl, State),
    Title = get_object_value("title", Bug),
    Notes = get_object_value("tags", Bug) ++
            get_bug_task_notes(get_object_value("entries", BugTasks)) ++
            ["https://launchpad.net/bugs/" ++ Number],
    {ok, Pid} = mup_server_sup:find_server(Message#message.server),
    mup_server:send(Pid, io_lib:fwrite("PRIVMSG ~s :Bug #~s: ~s ~s",
                                       [lists:nth(1, Message#message.params),
                                        Number, Title, format_notes(Notes)])).


get_bug_task_notes([Entry|Rest]) ->
    [erlang:binary_to_list(get_object_value("bug_target_display_name", Entry))
     ++ ":" ++
     get_bug_task_status(Entry)
     | get_bug_task_notes(Rest)];
get_bug_task_notes([]) ->
    [].


get_bug_task_status(Entry) ->
    Status = binary_to_list(get_object_value("status", Entry)),
    case get_object_value("assignee_link", Entry) of
        null ->
            Status;
        Url ->
            [_, Nick] = string:tokens(binary_to_list(Url), "~"),
            case Status of
                "New" ->
                    Status ++ " for " ++ Nick;
                "Confirmed" ->
                    Status ++ " for " ++ Nick;
                _ ->
                    Status ++ " by " ++ Nick
            end
    end.


perform_launchpad_request(UrlSuffix, State) ->
    Plugin = State#state.plugin,
    {MegaSecs, Secs, MicroSecs} = erlang:now(),
    random:seed(MegaSecs, Secs, MicroSecs),
    OAuthAccessToken = mup_util:option(oauth_access_token,
                                       Plugin#plugin.options),
    OAuthSecretToken = mup_util:option(oauth_secret_token,
                                       Plugin#plugin.options),
    OAuthNonce = integer_to_list(random:uniform(9999999)),
    OAuthTimestamp = integer_to_list(MegaSecs * 1000000 + Secs),
    Url = lists:flatten(["https://api.edge.launchpad.net/beta/", UrlSuffix]),
    AuthValue = ["OAuth realm=\"https://api.launchpad.net\", ",
                 "oauth_consumer_key=\"mup\", ",
                 "oauth_signature_method=\"PLAINTEXT\", ",
                 "oauth_token=\"", OAuthAccessToken, "\", ",
                 "oauth_signature=\"&", OAuthSecretToken, "\", ",
                 "oauth_nonce=\"", OAuthNonce, "\", ",
                 "oauth_timestamp=\"", OAuthTimestamp, "\""],
    Result = http:request(get, {Url, [{"Authorization", AuthValue}]}, [], []),
    io:format("Result: ~p~n", [Result]),
    {ok, {{_Http, 200, _Ok}, _Headers, Body}} = Result,
    {ok, Object, _Rest} = rfc4627:decode(Body),
    io:format("Object: ~p~n", [Object]),
    Object.


get_object_value(Key, [{Key, Value}|_]) ->
    Value;
get_object_value(Key, [_|Pairs]) ->
    get_object_value(Key, Pairs);
get_object_value(Key, {obj, Pairs}) ->
    get_object_value(Key, Pairs);
get_object_value(_Key, []) ->
    not_found.

get_object_value_test_() ->
    [?_assertEqual("value",
                   get_object_value("key", {obj, [{"key", "value"}]})),
     ?_assertEqual("value",
                   get_object_value("key", {obj, [{"key", "value"},
                                                  {"other", "other"}]})),
     ?_assertEqual("value",
                   get_object_value("key", {obj, [{"other", "other"},
                                                  {"key", "value"}]})),
     ?_assertEqual(not_found,
                   get_object_value("non-existent", {obj, [{"other", "other"},
                                                           {"key", "value"}]}))
    ].


format_notes(Notes) ->
    list_to_binary(["<", mup_util:join(Notes, "> <"), ">"]).

format_notes_test_() ->
    [?_assertEqual(<<"<abc>">>,
                   format_notes(["abc"])),
     ?_assertEqual(<<"<abc> <def>">>,
                   format_notes(["abc", "def"])),
     ?_assertEqual(<<"<abc> <def>">>,
                   format_notes([<<"abc">>, <<"def">>]))
    ].

get_option(Key, State) ->
    mup_util:option(Key, (State#state.plugin)#plugin.options).

get_option(Key, Default, State) ->
    mup_util:option(Key, (State#state.plugin)#plugin.options, Default).
