-module(mup_plugin_sup).
-behaviour(supervisor).
-export([start_link/0, init/1, start_plugin/1, find_plugin/1, get_plugins/0,
         handle_message/1, reconfig/0]).
-include("mup.hrl").


start_link() ->
    supervisor:start_link({local, ?MODULE}, ?MODULE, []).

init(_) ->
    {ok, {{one_for_one, 3, 60000}, []}}.

start_plugin(Name) ->
    {ok, Plugin} = mup_config:get_plugin(Name),
    io:format("Starting plugin \"~w\" (module ~w)~n",
              [Name, Plugin#plugin.module]),
    {ok, _Pid} = supervisor:start_child(?MODULE,
                                        {Name,
                                         {Plugin#plugin.module,
                                          start_link, [Name]},
                                         transient, 60, worker,
                                         [Plugin#plugin.module]}).


find_plugin(#plugin{}=Server) ->
    find_plugin(Server#plugin.name);
find_plugin(Name) ->
    Children = supervisor:which_children(?MODULE),
    find_plugin(Name, Children).

find_plugin(Name, [{Name, Pid, _Type, _Modules}|_Rest]) when is_pid(Pid) ->
    {ok, Pid};
find_plugin(Name, [_|Rest]) ->
    find_plugin(Name, Rest);
find_plugin(_Name, []) ->
    undefined.


get_plugins() ->
    get_plugins(supervisor:which_children(?MODULE)).

get_plugins([{Name, Pid, _Type, _Modules}|Rest]) when is_pid(Pid) ->
    [{Name, Pid}|get_plugins(Rest)];
get_plugins([_|Rest]) ->
    get_plugins(Rest);
get_plugins([]) ->
    [].


handle_message(Message) ->
    handle_message(get_plugins(), Message).

handle_message([{_Name, Pid}|Rest], Message) ->
    mup_plugin:handle_message(Pid, Message),
    handle_message(Rest, Message);
handle_message([], _Message) ->
    ok.


reconfig() ->
    WantedNames = mup_config:get_plugin_names(),
    reconfig(WantedNames),
    terminate_unwanted(get_plugins(), WantedNames).

reconfig([Name|Rest]) ->
    case find_plugin(Name) of
        undefined ->
            start_plugin(Name);
        {ok, Pid} ->
            mup_plugin:reconfig(Pid)
    end,
    reconfig(Rest);
reconfig([]) ->
    ok.

terminate_unwanted([{Name, _Pid}|Rest], WantedNames) ->
    case lists:member(Name, WantedNames) of
        true ->
            ok;
        false ->
            io:format("Stopping plugin \"~w\"~n", [Name]),
            supervisor:terminate_child(?MODULE, Name),
            supervisor:delete_child(?MODULE, Name)
    end,
    terminate_unwanted(Rest, WantedNames);
terminate_unwanted([], _WantedNames) ->
    ok.
