-module(mup_server).
-include("mup.hrl").
-include_lib("eunit/include/eunit.hrl").
-export([start_link/1, stop/1, send/2, network_module/1, reconfig/1]).
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).


-record(state,
        {server=#server{},
         status=disconnected,
         network_module,
         socket,
         nick,
         wanted_nick,
         channels=[],
         wanted_channels=[],
         ping_timer,
         last_pong,
         trace=true}).


start_link(Name) ->
    gen_server:start_link(?MODULE, Name, []).

send(Pid, Data) ->
    gen_server:cast(Pid, {send, Data}).

reconfig(Pid) ->
    gen_server:cast(Pid, reconfig).

stop(Pid) ->
    gen_server:call(Pid, stop).


network_module(Server) ->
    ServerOptions = Server#server.options,
    case mup_util:option(network_module, ServerOptions) of
        undefined ->
            case mup_util:option(ssl, ServerOptions) of
                true ->
                    ssl:start(), ssl;
                _ ->
                    gen_tcp
            end;
        Module ->
            Module
    end.

network_module_test_() ->
    [?_assertEqual(gen_tcp,
                   mup_server:network_module(#server{})),
     ?_assertEqual(ssl,
                   mup_server:network_module(#server{options=[ssl]})),
     ?_assertEqual(true,
                   mup_util:is_application_loaded(ssl)),
     ?_assertEqual(mock_tcp,
                   mup_server:network_module(#server{options=[{network_module,
                                                               mock_tcp}]}))
    ].

set_state_server(Server, State) ->
    WantedNick = mup_util:option(nick, Server#server.options, "mup"),
    WantedChannels = mup_util:option(channels, Server#server.options, []),
    State#state{server=Server,
                wanted_nick=WantedNick,
                wanted_channels=ordsets:from_list(WantedChannels)}.

schedule_ping(State) ->
    case State#state.ping_timer of
        undefined ->
            ok;
        Timer ->
            erlang:cancel_timer(Timer)
    end,
    State#state{ping_timer=erlang:send_after(30000, self(), ping)}.


init(Name) ->
    process_flag(trap_exit, true),
    % We message ourselves so that we can connect later at our own leisure.
    % This will free mup to do something else while we connect.
    self() ! connect,
    {ok, Server} = mup_config:get_server(Name),
    State = #state{network_module=network_module(Server)},
    {ok, set_state_server(Server, State)}.

handle_cast({send, Data}, State) ->
    send_data(State, Data),
    {noreply, State};

handle_cast(reconfig, State) ->
    self() ! ping, % Synchronize the state earlier.
    case mup_config:get_server((State#state.server)#server.name) of
        {ok, Server} ->
            {noreply, set_state_server(Server, State)};
        undefined ->
            {stop, normal, State}
    end.

handle_call(stop, _From, State) ->
    {stop, normal, ok, State}.


handle_info(connect, State) ->
    NetworkModule = State#state.network_module,
    {ok, Socket} = NetworkModule:connect(server_option(State, host),
                                         server_option(State, port),
                                         [{packet, line}]),
    NewState = State#state{status=connected,
                           socket=Socket,
                           last_pong=timestamp()},
    ok = send_command(password, NewState),
    ok = send_command(nick, NewState),
    ok = send_command(user, NewState),
    {noreply, schedule_ping(NewState)};

handle_info(ping, State) ->
    Now = timestamp(),
    if
        State#state.last_pong < Now - 60 ->
            trace("Ping timeout!", [], State),
            exit(ping_timeout);
        true ->
            ok
    end,
    send_command(ping, State),
    % Synchronize nick and channels.  These commands won't send
    % anything unless necessary to conform to the wanted state.
    send_command(nick, State),
    send_command(join, State),
    send_command(part, State),
    {noreply, schedule_ping(State)};

handle_info({tcp, _Socket, Data}, State) ->
    {ok, NewState} = process_message(State, Data),
    {noreply, NewState};

handle_info({ssl, _Socket, Data}, State) ->
    {ok, NewState} = process_message(State, Data),
    {noreply, NewState}.


terminate(_Reason, State) ->
    NetworkModule = State#state.network_module,
    case NetworkModule of
        mock_tcp ->
            ok;
        _ ->
            send_command(quit, State),
            NetworkModule:close(State#state.socket)
    end,
    ignored.

code_change(_OldVsn, State, _Extra) ->
    {ok, State}.


process_message(State, Data) ->
    Message = parse_message(Data),
    if
        Message#message.command /= 'PONG' ->
            [NoEOLData] = string:tokens(Data, "\r\n"),
            trace("Received: ~s", [NoEOLData], State);
        true ->
            ok
    end,
    {ok, NewState} = handle_command(Message#message.command, Message, State),
    mup_plugin_sup:handle_message(Message#message{server=NewState#state.server,
                                  mup_nick=NewState#state.nick}),
    {ok, NewState}.


handle_command('001', _Message, State) ->
    % We've got a welcome message. We're in!  In addition to keeping
    % track of it, we reset the nick to the one we actually want
    % (we may have opted to some variation to get in), and also send
    % join commands.
    NewState = State#state{status=registered,
                           nick=State#state.wanted_nick,
                           wanted_nick=server_option(State, nick, "mup")},
    send_command(join, NewState),
    {ok, NewState};

handle_command('PING', Message, State) ->
    send_data(State, "PONG " ++ lists:last(Message#message.params)),
    {ok, State};

handle_command('PONG', _Message, State) ->
    {ok, State#state{last_pong=timestamp()}};

handle_command('433', _Message, State)
        when State#state.status == connected ->
    % The nick we've sent is already in use, and we're not
    % yet registered. Let's try something else to get in.
    NewState = State#state{wanted_nick=State#state.wanted_nick ++ "_"},
    send_command(nick, NewState),
    {ok, NewState};

handle_command('NICK', Message, State)
        when State#state.nick == Message#message.nick ->
    % Our nick just changed!
    Nick = Message#message.text,
    trace("Nick changed to \"~s\"", [Nick], State),
    {ok, State#state{nick=Nick}};

handle_command('JOIN', Message, State)
        when State#state.nick == Message#message.nick ->
    % We've just joined a channel.
    Channel = Message#message.text,
    trace("Joined channel \"~s\".", [Channel], State),
    NewChannels = ordsets:add_element(Channel, State#state.channels),
    {ok, State#state{channels=NewChannels}};

handle_command('PART', Message, State)
        when State#state.nick == Message#message.nick ->
    % We've just left a channel.
    Channel = lists:nth(1, Message#message.params),
    trace("Left channel \"~s\".", [Channel], State),
    NewChannels = ordsets:del_element(Channel, State#state.channels),
    {ok, State#state{channels=NewChannels}};

handle_command(_Command, _Message, State) ->
    {ok, State}.


send_command(password, State) ->
    case server_option(State, password) of
        undefined ->
            ok;
        Password ->
            send_data(State, "PASS " ++ Password)
    end;

send_command(nick, State) ->
    if 
        State#state.wanted_nick == State#state.nick ->
            ok;
        true ->
            send_data(State, "NICK " ++ State#state.wanted_nick)
    end;

send_command(join, State) ->
    MissingChannels = ordsets:subtract(State#state.wanted_channels,
                                       State#state.channels),
    lists:foreach(fun (Channel) -> send_data(State, "JOIN " ++ Channel) end,
                  MissingChannels);

send_command(part, State) ->
    LeftoverChannels = ordsets:subtract(State#state.channels,
                                        State#state.wanted_channels),
    lists:foreach(fun (Channel) -> send_data(State, "PART " ++ Channel) end,
                  LeftoverChannels);

send_command(user, State) ->
    send_data(State, "USER mup 0 0 :Mup Pet");

send_command(ping, State) ->
    send_data(State#state{trace=false}, "PING :Hullah!");

send_command(quit, State) ->
    send_data(State, "QUIT :Coffee time, brb!").


send_data(State, Data) ->
    NetworkModule = State#state.network_module,
    Socket = State#state.socket,
    trace("Sending: ~s", [Data], State),
    NetworkModule:send(Socket, Data ++ "\r\n"),
    ok.
            

send_command_nick_test() ->
    State = #state{network_module=mock_tcp,
                   socket=the_socket,
                   wanted_nick="mopous"},
    spawn(fun () -> ok = send_command(nick, State) end),
    ?assertEqual({mock_tcp, send, the_socket, "NICK mopous\r\n"}, mock:next()),
    mock:answer(ok).

send_command_nick_when_already_got_it_test() ->
    State = #state{network_module=mock_tcp,
                   socket=the_socket,
                   nick="mopous",
                   wanted_nick="mopous"},
    spawn(fun () -> ok = send_command(nick, State) end),
    ?assertEqual(mock_next_timeout, mock:next()). % Nothing sent.

send_command_password_without_password_test() ->
    State = #state{network_module=mock_tcp,
                   socket=the_socket},
    spawn(fun () -> ok = send_command(password, State) end),
    ?assertEqual(mock_next_timeout, mock:next()). % No password should be sent.

send_command_password_test() ->
    Server = #server{options=[{password, "the password"}]},
    State = #state{server=Server,
                   network_module=mock_tcp,
                   socket=s},
    spawn(fun () -> ok = send_command(password, State) end),
    ?assertEqual({mock_tcp, send, s, "PASS the password\r\n"}, mock:next()),
    mock:answer(ok).

send_command_join_test() ->
    State = #state{network_module=mock_tcp,
                   socket=the_socket,
                   wanted_channels=["#a", "#b", "#c", "#d"],
                   channels=["#b", "#c"]},
    spawn(fun () -> ok = send_command(join, State) end),
    ?assertEqual({mock_tcp, send, the_socket, "JOIN #a\r\n"}, mock:next()),
    mock:answer(ok),
    ?assertEqual({mock_tcp, send, the_socket, "JOIN #d\r\n"}, mock:next()),
    mock:answer(ok),
    ?assertEqual(mock_next_timeout, mock:next()). % And nothing else.

send_command_part_test() ->
    State = #state{network_module=mock_tcp,
                   socket=the_socket,
                   wanted_channels=["#b", "#c"],
                   channels=["#a", "#b", "#c", "#d"]},
    spawn(fun () -> ok = send_command(part, State) end),
    ?assertEqual({mock_tcp, send, the_socket, "PART #a\r\n"}, mock:next()),
    mock:answer(ok),
    ?assertEqual({mock_tcp, send, the_socket, "PART #d\r\n"}, mock:next()),
    mock:answer(ok),
    ?assertEqual(mock_next_timeout, mock:next()). % And nothing else.

send_command_user_test() ->
    State = #state{network_module=mock_tcp,
                   socket=s},
    spawn(fun () -> ok = send_command(user, State) end),
    ?assertEqual({mock_tcp, send, s, "USER mup 0 0 :Mup Pet\r\n"}, mock:next()),
    mock:answer(ok).

send_command_ping_test() ->
    State = #state{network_module=mock_tcp,
                   socket=s},
    spawn(fun () -> ok = send_command(ping, State) end),
    ?assertEqual({mock_tcp, send, s, "PING :Hullah!\r\n"}, mock:next()),
    mock:answer(ok).

send_command_quit_test() ->
    State = #state{network_module=mock_tcp,
                   socket=s},
    spawn(fun () -> ok = send_command(quit, State) end),
    ?assertEqual({mock_tcp, send, s, "QUIT :Coffee time, brb!\r\n"},
                 mock:next()),
    mock:answer(ok).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Message parsing
%
% These functions will parse a message line into the message record.
%

parse_message([$:|Rest]) ->
    parse_message(prefix, Rest, [], #message{});
parse_message(Any) ->
    parse_message(command, Any, [], #message{}).

parse_message(prefix, [$ |Rest], Prefix, Message) ->
    ParsedPrefixMessage = parse_prefix(nick, Prefix, [], Message),
    parse_message(command, string:strip(Rest, left), [],
                  ParsedPrefixMessage#message{prefix=Prefix});
parse_message(prefix, [Char|Rest], Prefix, Message) ->
    parse_message(prefix, Rest, Prefix ++ [Char], Message);

parse_message(command, "\r\n", Command, Message) ->
    Message#message{command=erlang:list_to_atom(Command)};
parse_message(command, [$ |Rest], Command, Message) ->
    parse_message(params, string:strip(Rest, left), [],
                  Message#message{command=erlang:list_to_atom(Command)});
parse_message(command, [Char|Rest], Command, Message) ->
    parse_message(command, Rest, Command ++ [Char], Message);

parse_message(params, "\r\n", Param, Message) ->
    Message#message{params=Message#message.params ++ [Param]};
parse_message(params, [$ |Rest], Param, Message) ->
    parse_message(params, string:strip(Rest, left), [],
                  Message#message{params=Message#message.params ++ [Param]});
parse_message(params, [$:|Rest], [], Message) ->
    parse_message(text_param, string:strip(Rest, left), [], Message);
parse_message(params, [Char|Rest], Params, Message) ->
    parse_message(params, Rest, Params ++ [Char], Message);

parse_message(text_param, "\r\n", Text, Message) ->
    Param = [$:|Text],
    Message#message{params=Message#message.params ++ [Param],
                    text=Text};
parse_message(text_param, [Char|Rest], Text, Message) ->
    parse_message(text_param, Rest, Text ++ [Char], Message).

parse_prefix(nick, [], _Nick, Message) ->
    % It's a plain prefix, not in nick!user@host format.
    Message;
parse_prefix(nick, [$!|Rest], Nick, Message) ->
    parse_prefix(username, Rest, [], Message#message{nick=Nick});
parse_prefix(nick, [Char|Rest], Nick, Message) ->
    parse_prefix(nick, Rest, Nick ++ [Char], Message);

parse_prefix(username, [$@|Rest], Username, Message) ->
    Message#message{username=Username, userhost=Rest};
parse_prefix(username, [Char|Rest], Username, Message) ->
    parse_prefix(username, Rest, Username ++ [Char], Message).


parse_message_test_() ->
    [?_assertEqual(#message{command='CMD'},
                   parse_message("CMD\r\n")),
     ?_assertEqual(#message{prefix="prefix", command='CMD'},
                   parse_message(":prefix CMD\r\n")),
     ?_assertEqual(#message{prefix="prefix", command='CMD', params=["Yo!"]},
                   parse_message(":prefix CMD Yo!\r\n")),
     ?_assertEqual(#message{prefix="prefix", command='CMD',
                            params=["Hi", "there!"]},
                   parse_message(":prefix CMD Hi there!\r\n")),
     ?_assertEqual(#message{prefix="prefix", command='CMD', params=["Yo!"]},
                   parse_message(":prefix  CMD  Yo!\r\n")),
     ?_assertEqual(#message{command='CMD', params=[":Some text"],
                            text="Some text"},
                   parse_message("CMD :Some text\r\n")),
     ?_assertEqual(#message{command='CMD', params=["par:am", ":Some text"],
                            text="Some text"},
                   parse_message("CMD par:am :Some text\r\n")),
     ?_assertEqual(#message{prefix="nick!user@host", command='CMD',
                            nick="nick", username="user", userhost="host"},
                   parse_message(":nick!user@host CMD\r\n"))
    ].


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Some utility functions
%

server_option(State, Name) ->
    mup_util:option(Name, (State#state.server)#server.options).

server_option(State, Name, Default) ->
    mup_util:option(Name, (State#state.server)#server.options, Default).

server_option_test_() ->
    [?_assertEqual(bar,
                   server_option(#state{server=#server{options=[{foo, bar}]}},
                                 foo)),
     ?_assertEqual(undefined,
                   server_option(#state{server=#server{options=[{foo, bar}]}},
                                 bar)),
     ?_assertEqual(bar,
                   server_option(#state{server=#server{options=[{foo, bar}]}},
                                 foo, default)),
     ?_assertEqual(default,
                   server_option(#state{server=#server{options=[{foo, bar}]}},
                                 bar, default))
    ].


timestamp() ->
    {MegaSecs, Secs, _MicroSecs} = erlang:now(),
    MegaSecs * 1000000 + Secs.


trace(Format, Args, State) ->
    case State#state.trace of
        true ->
            io:format("[~s] " ++ Format ++ "~n",
                      [(State#state.server)#server.name|Args]);
        false ->
            ok
    end.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% General integration tests
%

mock_until_connected() ->
    mock_until_connected([]).

mock_until_connected("USER" ++ _) ->
    ok;
mock_until_connected(_) ->
    case mock:next() of
        {mock_tcp, connect, _, _, _} ->
            mock:answer({ok, the_socket}),
            mock_until_connected([]);
        {mock_tcp, send, _, Data} ->
            mock:answer(ok),
            mock_until_connected(Data)
    end.

mup_server_test_() ->
    {foreach,
     fun () ->
        {ok, ConfigPid} = mup_config:start_link(),
        {ok, ServerSupPid} = mup_server_sup:start_link(),
        {ok, PluginSupPid} = mup_plugin_sup:start_link(),
        {ConfigPid, ServerSupPid, PluginSupPid}
     end,
     fun ({ConfigPid, ServerSupPid, PluginSupPid}) ->
        mup_util:sync_stop(ConfigPid),
        mup_util:sync_stop(ServerSupPid),
        mup_util:sync_stop(PluginSupPid)
     end,
     [fun basic_connection_subtest/0,
      fun connection_with_server_password_subtest/0,
      fun send_subtest/0,
      fun ping_subtest/0]}.

basic_connection_subtest() ->
    mup_config:add_term(#server{name=example,
                                options=[{host, "irc.example.com"},
                                         {port, 6667},
                                         {network_module, mock_tcp}]}),
    {ok, _Pid} = mup_server_sup:start_server(example),

    ?assertEqual({mock_tcp, connect, "irc.example.com", 6667, [{packet, line}]},
                 mock:next()),
    mock:answer({ok, the_socket}),

    ?assertEqual({mock_tcp, send, the_socket, "NICK mup\r\n"},
                 mock:next()),
    mock:answer(ok),

    ?assertEqual({mock_tcp, send, the_socket, "USER mup 0 0 :Mup Pet\r\n"},
                 mock:next()),
    mock:answer(ok).

connection_with_server_password_subtest() ->
    mup_config:add_term(#server{name=example,
                                options=[{host, "irc.example.com"},
                                         {port, 6667},
                                         {password, "the password"},
                                         {network_module, mock_tcp}]}),
    {ok, _Pid} = mup_server_sup:start_server(example),

    ?assertEqual({mock_tcp, connect, "irc.example.com", 6667, [{packet, line}]},
                 mock:next()),
    mock:answer({ok, the_socket}),
    ?assertEqual({mock_tcp, send, the_socket, "PASS the password\r\n"},
                 mock:next()),
    mock:answer(ok),
    ?assertEqual({mock_tcp, send, the_socket, "NICK mup\r\n"},
                 mock:next()),
    mock:answer(ok),
    ?assertEqual({mock_tcp, send, the_socket, "USER mup 0 0 :Mup Pet\r\n"},
                 mock:next()),
    mock:answer(ok).

send_subtest() ->
    mup_config:add_term(#server{name=example,
                                options=[{host, "irc.example.com"},
                                         {port, 6667},
                                         {network_module, mock_tcp}]}),
    {ok, Pid} = mup_server_sup:start_server(example),
    mock_until_connected(),
    mup_server:send(Pid, "THE COMMAND"),
    ?assertEqual({mock_tcp, send, the_socket, "THE COMMAND\r\n"}, mock:next()),
    mock:answer(ok).

ping_subtest() ->
    mup_config:add_term(#server{name=example,
                                options=[{host, "irc.example.com"},
                                         {port, 6667},
                                         {network_module, mock_tcp}]}),
    {ok, Pid} = mup_server_sup:start_server(example),
    mock_until_connected(),
    Pid ! {tcp, the_socket, "PING :Hello!\r\n"},
    ?assertEqual({mock_tcp, send, the_socket, "PONG :Hello!\r\n"}, mock:next()),
    mock:answer(ok).
